/*******************************************************************************
 *
 * File Name: THmain.c
 * Author: Richard Baxter
 * Project: Advanced Neural Network (ANN) - Fractal Based Neural Network
 * Project Version: 1b13d 29-Sept-07
 * Comments:
 *
 *******************************************************************************/

#include <ctime>
#include <cstdlib>	//for randon number generation
#include <cmath>

#include <string>
#include <iostream>
#include <fstream>
using namespace std;


#include "neuralNetworkFormation.h"
#include "neuralNetworkTraining.h"
#include "THneuralNetworkParser.h"


#include "THData.h"
#include "THmain.h"

int main()
{
	//parse test harness Data set
	THparseTestDataFile();

	int maxFolds = MAX_NUM_FOLDS_TH;		//maximum number of folds are same regardless of TEST_LEVEL_X_TH

	//Neural Network initialisations
	Neuron * firstInputNeuron = new Neuron();		//only the frontNeuronReferences of the firstInputNeuron are used - they are used in establishment of neuron relationships
	Neuron * firstOutputNeuron;

	long numberOfInputNeurons = NUM_INPUT_NEURONS_TH;
	long numberOfOutputNeurons = NUM_OUTPUT_NEURONS_TH;

	//form neural network based upon Test Harness Data set
#ifdef TEST_LEVEL_0_TH

	cout << "TEST_LEVEL_0_TH \n" <<endl;
	cout << "\n" << "Number of Hidden Neurons = " << numberOfHiddenNeurons << "\n" <<endl;
	#define NUMBER_OF_HIDDEN_NEURONS_HEURISTIC_TH (1.0F)

	firstOutputNeuron = new Neuron();
	long meanNumberOfHiddenLayerNeurons = long(((numberOfInputNeurons+numberOfOutputNeurons)/2.0)*NUMBER_OF_HIDDEN_NEURONS_HEURISTIC_TH);
	formSimpleNeuralNet(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, meanNumberOfHiddenLayerNeurons, numberOfOutputNeurons);
	THTrainNeuralNetwork(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, maxFolds);

#elif defined TEST_LEVEL_1_TH


	/*
	network structure

	5 input neurons
	X hidden neurons
	3 output neurons

	each neuron of a layer is connected to each other neuron of the adjacent layer(s)
	*/

	int hiddenNeuronHeuristic = int(0.5F*(float(numberOfInputNeurons)+float(numberOfOutputNeurons)));
		//heuristic Calculated for Number Of Hidden Neurons

	int numberOfHiddenNeurons;

	float division;

	//#define ONLY_USE_HIDDEN_NEURON_HEURISTIC
	#ifdef ONLY_USE_HIDDEN_NEURON_HEURISTIC
	division = 1;
	//change hiddenNeuronHeuristic to the numberOfHiddensNeurons being timed
	//hiddenNeuronHeuristic = 7;
	#else
	division = 2;
	#endif

	for(numberOfHiddenNeurons = int(float(hiddenNeuronHeuristic)/division); numberOfHiddenNeurons <= int(float(hiddenNeuronHeuristic)*division); numberOfHiddenNeurons++)
	{
		cout << "TEST_LEVEL_1_TH \n" <<endl;
		cout << "\n" << "Number of Hidden Neurons = " << numberOfHiddenNeurons << "\n" <<endl;
		firstOutputNeuron = new Neuron();
		formSimpleNeuralNet(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfHiddenNeurons, numberOfOutputNeurons);
		THTrainNeuralNetwork(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, maxFolds);
	}

#elif defined TEST_LEVEL_2_TH

	long meanNumberOfHiddenLayers = ADV_MEAN_NUMBER_OF_HIDDEN_LAYERS_TH;
	long meanSubnetDivergenceFactor = MEAN_SUBNET_DIVERGENCE_FACTOR_TH;

	cout << "TEST_LEVEL_2_TH \n" <<endl;
	cout << "\n" << "Mean of Hidden Layers = " << meanNumberOfHiddenLayers <<endl;
	cout << "Mean Subnet Divergence Factor = " << meanSubnetDivergenceFactor << "\n" <<endl;

	firstOutputNeuron = formNeuralNet(firstInputNeuron, numberOfInputNeurons, numberOfOutputNeurons, meanNumberOfHiddenLayers, IRRELEVANT);
	THTrainNeuralNetwork(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, maxFolds);

#elif defined TEST_LEVEL_3_TH

	for(long meanNumberOfHiddenLayers = ADV_MEAN_NUMBER_OF_HIDDEN_LAYERS_TH/2; meanNumberOfHiddenLayers < ADV_MEAN_NUMBER_OF_HIDDEN_LAYERS_TH*2; meanNumberOfHiddenLayers++)
	{
		for(long meanSubnetDivergenceFactor = MEAN_SUBNET_DIVERGENCE_FACTOR_TH/2; meanSubnetDivergenceFactor < MEAN_SUBNET_DIVERGENCE_FACTOR_TH*2; meanSubnetDivergenceFactor++)
		{
			cout << "TEST_LEVEL_3_TH \n" <<endl;
			cout << "\n" << "Mean of Hidden Layers = " << meanNumberOfHiddenLayers <<endl;
			cout << "Mean Subnet Divergence Factor = " << meanSubnetDivergenceFactor << "\n" <<endl;
			firstOutputNeuron = formNeuralNet(firstInputNeuron, numberOfInputNeurons, numberOfOutputNeurons, meanNumberOfHiddenLayers, IRRELEVANT);
			THTrainNeuralNetwork(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, maxFolds);
		}
	}


#elif defined TEST_LEVEL_4_TH

	long meanNumberOfHiddenLayers = ADV_MEAN_NUMBER_OF_HIDDEN_LAYERS_TH;
	long meanSubnetDivergenceFactor = MEAN_SUBNET_DIVERGENCE_FACTOR_TH;
	long maxNumRecursiveSubnets = MAX_NUM_RECURSIVE_SUBNETS_TH;

	cout << "TEST_LEVEL_4_TH \n" <<endl;
	cout << "\n" << "Mean of Hidden Layers = " << meanNumberOfHiddenLayers <<endl;
	cout << "Mean Subnet Divergence Factor = " << meanSubnetDivergenceFactor <<endl;
	cout << "Max Number Recursive Subnets = " << maxNumRecursiveSubnets << "\n" <<endl;

	firstOutputNeuron = formANN(firstInputNeuron, numberOfInputNeurons, numberOfOutputNeurons, meanNumberOfHiddenLayers, meanSubnetDivergenceFactor, maxNumRecursiveSubnets);
	THTrainNeuralNetwork(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, maxFolds);

#elif defined TEST_LEVEL_5_TH


	for(long meanNumberOfHiddenLayers = ADV_MEAN_NUMBER_OF_HIDDEN_LAYERS_TH/2; meanNumberOfHiddenLayers < ADV_MEAN_NUMBER_OF_HIDDEN_LAYERS_TH*2; meanNumberOfHiddenLayers++)
	{
		for(long meanSubnetDivergenceFactor = MEAN_SUBNET_DIVERGENCE_FACTOR_TH/2; meanSubnetDivergenceFactor < MEAN_SUBNET_DIVERGENCE_FACTOR_TH*2; meanSubnetDivergenceFactor++)
		{
			for(long maxNumRecursiveSubnets = MAX_NUM_RECURSIVE_SUBNETS_TH/2; maxNumRecursiveSubnets < MAX_NUM_RECURSIVE_SUBNETS_TH*2; maxNumRecursiveSubnets++)
			{
				cout << "TEST_LEVEL_5_TH \n" <<endl;
				cout << "\n" << "Mean of Hidden Layers = " << meanNumberOfHiddenLayers <<endl;
				cout << "Mean Subnet Divergence Factor = " << meanSubnetDivergenceFactor <<endl;
				cout << "Max Number Recursive Subnets = " << maxNumRecursiveSubnets << "\n" <<endl;
				firstOutputNeuron = formANN(firstInputNeuron, numberOfInputNeurons, numberOfOutputNeurons, meanNumberOfHiddenLayers, meanSubnetDivergenceFactor, maxNumRecursiveSubnets);
				THTrainNeuralNetwork(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, maxFolds);
			}
		}
	}

#else
	cout << "error: no TEST_LEVEL_X_TH defined - recompilation required \n" <<endl;
#endif



}
