/*******************************************************************************
 *
 * File Name: THneuralNetworkParser.cpp
 * Author: Richard Baxter
 * Project: Advanced Neural Network (ANN) - Fractal Based Neural Network
 * Project Version: 1b13d 29-Sept-07
 * Comments:
 *
 *******************************************************************************/


#include <iostream>
#include <fstream>
#include <ctime>
#include <cstring>
#include <cstdlib>	//for randon number generation
#include <cmath>


#include <string>

using namespace std;





#include "THData.h"
#include "THneuralNetworkParser.h"



#define	SOME_REALLY_SMALL_FLOAT 0.0F
#define	SOME_REALLY_LARGE_FLOAT 500.0F

void THparseTestDataFile()
{
	char currentNumberString[MAX_CHARACTERS_TH] = "";

	//initialise shared variables
	maxSetsOfData = MAX_SETS_OF_DATA_TH;
	strcpy(dataFileName, DATA_FILE_NAME_TH);

	currentNumberString[0] = '\0';

	/* Data file layout

	1:	Class attribute (1 = normal, 2 = hyper, 3 = hypo)
	2:	T3-resin uptake test. (A percentage)
	3:	Total Serum thyroxin as measured by the isotopic
		displacement method.
	4: 	Total serum triiodothyronine as measured by radioimmuno
		assay.
	5: 	basal thyroid-stimulating hormone (TSH) as measured by
		radioimmuno assay.
	6: 	Maximal absolute difference of TSH value after injection of
		200 micro grams of thyrotropin-releasing hormone as compared
		to the basal value.

	*/

	int currentRow = 0;
	int currentColumn = 0;

	ifstream textFile(dataFileName);
	char c;

	//while ((*textFile).good())
	while ((textFile).get(c))
	{
		switch(c)
		{
			case '\n':
			{
				inputData[currentRow][currentColumn] = atof(currentNumberString);
				currentNumberString[0] = '\0';

					#ifdef DEBUG
					//cout << "DEBUG: inputData[" << currentRow << "][" << currentColumn << "] = " << inputData[currentRow][currentColumn] << endl;
					#endif

				currentRow++;
				currentColumn = 0;
				break;
			}
			case ',':
			{
				inputData[currentRow][currentColumn] = atof(currentNumberString);
				currentNumberString[0] = '\0';
				currentColumn++;
				break;
			}
			default:
			{
				char characterString[2];
				characterString[0] = c;
				characterString[1] = '\0';
				strcat(currentNumberString, characterString);
				break;
			}
		}
	}

	//now normalise data;


	//1. calculate max/min values

	float maxMin[MAX_INPUT_DATA_COLUMNS_TH][2];
		//stores maximum [0] and minimum [1] value for each type of inputData

	int q, i, j;
	for(q=0;q<MAX_INPUT_DATA_COLUMNS_TH;q++)
	{
		maxMin[q][0] = SOME_REALLY_SMALL_FLOAT; //0
		maxMin[q][1] = SOME_REALLY_LARGE_FLOAT;  //500
	}

	for(i=0;i<MAX_INPUT_DATA_ROWS_TH;i++)
	{
		for(j=0;j<MAX_INPUT_DATA_COLUMNS_TH;j++)
		{
			#ifdef DEBUG
			//cout << "DEBUG: inputData[" << i << "][" << j << "] = " << inputData[i][j] << endl;
			cout << "DEBUG: maxMin[" << j << "][1] = " << maxMin[j][1] << endl;
			#endif

			if(inputData[i][j] > maxMin[j][0])
			{
				maxMin[j][0] = inputData[i][j];
			}


			if(inputData[i][j] < maxMin[j][1])
			{
				maxMin[j][1] = inputData[i][j];

				#ifdef DEBUG
				/*
				if(inputData[i][j] < -100.0F)
				{
					cout << "DEBUG: error found inputData[" << i << "][" << j << "] = " << inputData[i][j] << endl;
				}
				*/
				#endif
			}
		}
	}


	#ifdef DEBUG
	for(j=0;j<MAX_INPUT_DATA_COLUMNS_TH;j++)
	{
		cout << "Max value of input type (column) = " << j << " = " << maxMin[j][0] << endl;
		cout << "Min value of input type (column) = " << j << " = " << maxMin[j][1] << endl;
	}
	#endif


	//2. insert normalised data entries
	for(i=0;i<MAX_INPUT_DATA_ROWS_TH;i++)
	{
		for(j=0;j<MAX_INPUT_DATA_COLUMNS_TH;j++)
		{
			if(j==0)
			{
				normalisedInputData[i][j] = inputData[i][j];
			}
			else
			{
				normalisedInputData[i][j] = ((inputData[i][j] - maxMin[j][1]) / (maxMin[j][0] - maxMin[j][1]));
			}
			#ifdef DEBUG
			//Testing
			/*
			cout << "DEBUG:inputData[i][j] = " << inputData[i][j] << endl;
			cout << "DEBUG:normalisedInputData[i][j] = " << normalisedInputData[i][j] << endl;
			cout << "DEBUG:maxMin[j][0] = " << maxMin[j][0] << endl;
			cout << "DEBUG: maxMin[j][1] = " << maxMin[j][1] << endl;
			cout << "\n" << endl;
			*/
			#endif
		}
	}



	//randomise normalisedInputData {so that classes are evenly distributed between rows)


	srand(static_cast<unsigned>(time(0)));

	bool stillToFill = true;

	for(int i=0;i < MAX_INPUT_DATA_ROWS_TH; i++)
	{
		normalisedInputDataTempFilled[i] = false;
	}

	int cnt = 0;

	while(stillToFill)
	{
		int res = (int) (float(rand() * float(MAX_INPUT_DATA_ROWS_TH))/(float(RAND_MAX)+1.0));
		//Testing
		//cout << "res = " << res << endl;

		if(normalisedInputDataTempFilled[res] == false)
		{//fill it
			for(int j=0;j<MAX_INPUT_DATA_COLUMNS_TH;j++)
			{
				//Testing
				//cout << "cnt = " << cnt << endl;

				normalisedInputDataTemp[res][j] = normalisedInputData[cnt][j];

			}
			normalisedInputDataTempFilled[res] = true;
			cnt++;
		}
		else
		{//do not fill it

		}

		//check if still entries to fill...
		stillToFill = false;
		for(int i=0;i<MAX_INPUT_DATA_ROWS_TH;i++)
		{
			if(normalisedInputDataTempFilled[i] == false)
			{
				stillToFill = true;
			}
			else
			{
				//do nothing
			}
		}
	}

	//now copy all normalisedInputDataTemp back into normalisedInputData
	for(int i=0;i<MAX_INPUT_DATA_ROWS_TH;i++)
	{
		for(int j=0;j<MAX_INPUT_DATA_COLUMNS_TH;j++)
		{
			normalisedInputData[i][j] = normalisedInputDataTemp[i][j];
			#ifdef DEBUG
			cout << "normalisedInputData[" << i << "][" << j << "] = " << normalisedInputData[i][j] << endl;
			#endif
		}
	}



	/*
	#ifdef DEBUG

	cout << "inputData[i][j] = " << inputData[0][1] << "\n" << endl;
	cout << "inputData[i][j] = " << inputData[0][2] << "\n" << endl;
	cout << "inputData[i][j] = " << inputData[0][3] << "\n" << endl;
	cout << "inputData[i][j] = " << inputData[0][4] << "\n" << endl;
	cout << "inputData[i][j] = " << inputData[0][5] << "\n" << endl;

	cout << "normalisedInputData[i][j] = " << normalisedInputData[0][1] << "\n" << endl;
	cout << "normalisedInputData[i][j] = " << normalisedInputData[0][2] << "\n" << endl;
	cout << "normalisedInputData[i][j] = " << normalisedInputData[0][3] << "\n" << endl;
	cout << "normalisedInputData[i][j] = " << normalisedInputData[0][4] << "\n" << endl;
	cout << "normalisedInputData[i][j] = " << normalisedInputData[0][5] << "\n" << endl;

	#endif
	*/

}
