/*******************************************************************************
 *
 * File Name: neuralNetworkFormation.cpp
 * Author: Richard Baxter
 * Project: Advanced Neural Network (ANN) - Fractal Based Neural Network
 * Project Version: 1b13d 29-Sept-07
 * Comments:
 *
 *******************************************************************************/

 //IMPORTANT CODING NOTE - on 10-dec-06/1b6b I have started to remove the neuronReference class as circular referencing does not appear to be allowed in C++
//NB when create NeuronList class change all referecnes to "...List->frontNeuronReferences.." to "...List->neuronReferences" as Neurons when declared for the purpose of creating neuron lists only have their frontNeuronReferences used

#include <ctime>
#include <cstdlib>	//for randon number generation
#include <cmath>

#include <string>
#include <iostream>
#include <fstream>
using namespace std;

#include "neuralNetworkFormation.h"


//const int numberOfInputNeurons = NUM_INPUT_NEURONS_TH;
//const int numberOfOutputNeurons = NUM_OUTPUT_NEURONS_TH;
//const int maxFolds = MAX_NUM_FOLDS_TH;
//const int maxSetsOfData = MAX_SETS_OF_DATA_TH;

static Neuron * formNondistinctLayeredNeuralNetwork(Neuron * firstInputNeuronInNetwork, long numberOfInputNeurons, long numberOfOutputNeurons, long meanNumberOfHiddenLayers, double meanSubnetDivergenceFactor, long maxNumberOfRecursiveSubnets, long currentNumberOfRecursiveSubnets);
	static Neuron * formNondistinctLayeredNeuralNetwork(Neuron * firstInputNeuronInNetwork, long numberOfInputNeurons, long numberOfOutputNeurons, long meanNumberOfHiddenLayers, double meanSubnetDivergenceFactor, long maxNumberOfRecursiveSubnets, long currentNumberOfRecursiveSubnets);
	static Neuron * fillNondistinctHiddenLayer(Neuron * firstInputNeuronInNetwork, Neuron * firstNeuronInCurrentLayer, long numberOfInputNeurons, long numberOfOutputNeurons, long currentNumberOfHiddenNeurons, long currentNumberOfHiddenLayers, long numberOfHiddenLayers, double meanSubnetDivergenceFactor, long maxNumberOfRecursiveSubnets, long currentNumberOfRecursiveSubnets);
		static long calculateNumberOfNeuronsInNewFrontLayer(long numberOfInputNeurons, long numberOfOutputNeurons, long currentNumberOfHiddenNeurons, long currentNumberOfHiddenLayers, long numberOfHiddenLayers, double meanDivergenceFactor);
			static double advancedDivergenceFactor(long currentNumberOfHiddenLayersFromNearestEndPoint, long numberOfHiddenLayers);
		static long calcMeanNumberOfHiddenLayersInSubnet(long numberOfInputNeurons, long numberOfOutputNeurons, double meanSubnetDivergenceFactor);



/************************************************************ Form Neural Network Routines ******************************************************/

/*forms a 3 layered net similar to that used in original c project*/
//assumes firstInputNeuron and firstOutputNeuron have already been filled
void formSimpleNeuralNet(Neuron * firstInputNeuron, Neuron * firstOutputNeuron, long numberOfInputNeurons, long numberOfHiddenNeurons, long numberOfOutputNeurons)
{
	/*NB simple neural network heursitc;
	number of hidden neurons = average of input and output neurons
	*/

	Neuron * firstHiddenNeuron = new Neuron();

	/*Create all Neurons in Simply Layered Net*/

	//initiate Input Layer in Neural Network
	Neuron * currentNeuronReference = firstInputNeuron;

	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		currentNeuronReference->firstNeuronInFrontLayer = firstHiddenNeuron;
		currentNeuronReference->hasFrontLayer = true;

		Neuron * newNeuron = new Neuron();
		currentNeuronReference->nextNeuron = newNeuron;
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}

	//initiate Hidden Layer in Neural Network
	currentNeuronReference = firstHiddenNeuron;
	for(long i = 0; i < numberOfHiddenNeurons; i++)
	{
		currentNeuronReference->firstNeuronInFrontLayer = firstOutputNeuron;
		currentNeuronReference->firstNeuronInBackLayer = firstInputNeuron;
		currentNeuronReference->hasFrontLayer = true;
		currentNeuronReference->hasBackLayer = true;

		Neuron * newNeuron = new Neuron();
		currentNeuronReference->nextNeuron = newNeuron;
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}

	//initiate Output Layer in Neural Network
	currentNeuronReference = firstOutputNeuron;
	for(long i = 0; i < numberOfOutputNeurons; i++)
	{
		currentNeuronReference->firstNeuronInBackLayer = firstHiddenNeuron;
		currentNeuronReference->hasBackLayer = true;

		Neuron * newNeuron = new Neuron();
		currentNeuronReference->nextNeuron = newNeuron;
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}


	/*Link all Neurons in Simply Layered Net - fill frontNeuronReferences and BackNeuronReferences for all neurons*/

	//Link ALL input and hidden layer neurons neurons (fill in frontNeuronReferences/BackNeuronReferences list for all relevant neurons)

	Neuron * currentBackLayerNeuron = firstInputNeuron;
	Neuron * currentForwardLayerNeuron = firstHiddenNeuron;
	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		for(long j = 0; j < numberOfHiddenNeurons; j++)
		{
			NeuronConnection * newNeuronConnection = new NeuronConnection();

			currentForwardLayerNeuron->numBackNeuronConnections = numberOfInputNeurons;
			currentBackLayerNeuron->numFrontNeuronConnections = numberOfHiddenNeurons;
			currentForwardLayerNeuron->backNeuronReferences[i] = currentBackLayerNeuron;
			currentBackLayerNeuron->frontNeuronReferences[j] = currentForwardLayerNeuron;
			currentForwardLayerNeuron->backNeuronRefConnections[i] = newNeuronConnection;
			currentBackLayerNeuron->frontNeuronRefConnections[j] = newNeuronConnection;

			currentForwardLayerNeuron = currentForwardLayerNeuron->nextNeuron;
		}
		currentBackLayerNeuron = currentBackLayerNeuron->nextNeuron;
	}


	//Link ALL hidden and output layer neurons neurons (fill in frontNeuronReferences/BackNeuronReferences list for all relevant neurons)
	currentBackLayerNeuron = firstHiddenNeuron;
	currentForwardLayerNeuron = firstOutputNeuron;
	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		for(long j = 0; j < numberOfHiddenNeurons; j++)
		{
			NeuronConnection * newNeuronConnection = new NeuronConnection();

			currentForwardLayerNeuron->numBackNeuronConnections = numberOfHiddenNeurons;
			currentBackLayerNeuron->numFrontNeuronConnections = numberOfOutputNeurons;
			currentForwardLayerNeuron->backNeuronReferences[i] = currentBackLayerNeuron;
			currentBackLayerNeuron->frontNeuronReferences[j] = currentForwardLayerNeuron;
			currentForwardLayerNeuron->backNeuronRefConnections[i] = newNeuronConnection;
			currentBackLayerNeuron->frontNeuronRefConnections[j] = newNeuronConnection;

			currentForwardLayerNeuron = currentForwardLayerNeuron->nextNeuron;
		}
		currentBackLayerNeuron = currentBackLayerNeuron->nextNeuron;
	}
}



















/************************************************************ End Form Neural Network Routines ******************************************************/



















	//this is currently only psuedo code

/************************************************************ Form Advanced Neural Network Routines ******************************************************/

Neuron * formNeuralNet(Neuron * firstInputNeuron, long numberOfInputNeurons, long numberOfOutputNeurons, long meanNumberOfHiddenLayers)
{
	Neuron * firstOutputNeuronInNetwork;

	double meanSubnetDivergenceFactor = 0.0;	/*no divergence = no ANN*/
	long maxNumberOfRecursiveSubnets = 0;		/*no recursive subnets = no ANN*/

	firstOutputNeuronInNetwork = formNondistinctLayeredNeuralNetwork(firstInputNeuron, numberOfInputNeurons, numberOfOutputNeurons, meanNumberOfHiddenLayers, meanSubnetDivergenceFactor, maxNumberOfRecursiveSubnets, 0);

	return firstOutputNeuronInNetwork;
}


#ifdef ANN
Neuron * formAdvancedNeuralNetwork(Neuron * firstInputNeuron, long numberOfInputNeurons, long numberOfOutputNeurons, long meanNumberOfHiddenLayers, double meanSubnetDivergenceFactor, long maxNumberOfRecursiveSubnets)
{
	Neuron * firstOutputNeuronInNetwork;

	firstOutputNeuronInNetwork = formNondistinctLayeredNeuralNetwork(firstInputNeuron, numberOfInputNeurons, numberOfOutputNeurons, meanNumberOfHiddenLayers, meanSubnetDivergenceFactor, maxNumberOfRecursiveSubnets, 0);

	return firstOutputNeuronInNetwork;
}
#endif

static Neuron * formNondistinctLayeredNeuralNetwork(Neuron * firstInputNeuronInNetwork, long numberOfInputNeurons, long numberOfOutputNeurons, long meanNumberOfHiddenLayers, double meanSubnetDivergenceFactor, long maxNumberOfRecursiveSubnets, long currentNumberOfRecursiveSubnets)
{
	Neuron * firstOutputNeuronInNetwork;

	//seed random number generator
	srand(static_cast<unsigned>(time(0)));


	//initiate Input Layer in Neural Network
	Neuron * currentNeuron = firstInputNeuronInNetwork;

	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		Neuron * newNeuron = new Neuron();
		currentNeuron->nextNeuron = newNeuron;
		currentNeuron = currentNeuron->nextNeuron;
	}

	firstOutputNeuronInNetwork = fillNondistinctHiddenLayer(firstInputNeuronInNetwork, firstInputNeuronInNetwork, numberOfInputNeurons, numberOfOutputNeurons, numberOfInputNeurons, 0, meanNumberOfHiddenLayers, meanSubnetDivergenceFactor, maxNumberOfRecursiveSubnets, currentNumberOfRecursiveSubnets);

	return firstOutputNeuronInNetwork;
}


static Neuron * fillNondistinctHiddenLayer(Neuron * firstInputNeuronInNetwork, Neuron * firstNeuronInCurrentLayer, long numberOfInputNeurons, long numberOfOutputNeurons, long currentNumberOfHiddenNeurons, long currentNumberOfHiddenLayers, long numberOfHiddenLayers, double meanSubnetDivergenceFactor, long maxNumberOfRecursiveSubnets, long currentNumberOfRecursiveSubnets)
{
	Neuron * firstOutputNeuronInNetwork;

	//Create neurons for the new front layer

	long numberOfNeuronsInNewFrontLayer;

	#define IRRELEVANT 0.0
	numberOfNeuronsInNewFrontLayer = calculateNumberOfNeuronsInNewFrontLayer(numberOfInputNeurons, numberOfOutputNeurons, currentNumberOfHiddenNeurons, currentNumberOfHiddenLayers, numberOfHiddenLayers, IRRELEVANT);

	Neuron * firstNeuronInNewFrontLayer = new Neuron;
	Neuron * currentNeuronReference = firstNeuronInNewFrontLayer;
	for(long i=0; i<numberOfNeuronsInNewFrontLayer; i++)
	{
			//fill in network description variables for this layer (CHECK THIS - this is incomplete)
		currentNeuronReference->firstNeuronInBackLayer = firstNeuronInCurrentLayer;
		currentNeuronReference->hasBackLayer = true;

		Neuron * newNeuron = new Neuron();
		currentNeuronReference->nextNeuron = newNeuron;
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}


		//fill in network description variables for this layer (CHECK THIS - this is incomplete)
	currentNeuronReference = firstNeuronInCurrentLayer;
	long numberOfNeuronsOnCurrentLayer = 0;
	while(currentNeuronReference -> nextNeuron != NULL)
	{
		currentNeuronReference->hasFrontLayer = true;
		currentNeuronReference->firstNeuronInFrontLayer = firstNeuronInNewFrontLayer;

		currentNeuronReference = currentNeuronReference->nextNeuron;
		numberOfNeuronsOnCurrentLayer++;
	}


	//Modify associations/links based upon layers of intermediate/hidden neurons

	//Link ALL current/back and front layer neurons neurons (fill in frontNeuronReferences/BackNeuronReferences list for all relevant neurons)
			//CHECK THIS - this can be changed in the future to link collections of neurons (not all)

	Neuron * currentBackLayerNeuron = firstNeuronInCurrentLayer;
	Neuron * currentForwardLayerNeuron = firstNeuronInCurrentLayer->firstNeuronInFrontLayer;
	for(long i = 0; i < numberOfNeuronsOnCurrentLayer; i++)
	{
		for(long j = 0; j < numberOfNeuronsInNewFrontLayer; j++)
		{
			NeuronConnection * newNeuronConnection = new NeuronConnection();

			currentForwardLayerNeuron->numBackNeuronConnections = numberOfNeuronsOnCurrentLayer;
			currentBackLayerNeuron->numFrontNeuronConnections = numberOfNeuronsInNewFrontLayer;
			currentForwardLayerNeuron->backNeuronReferences[i] = currentBackLayerNeuron;
			currentBackLayerNeuron->frontNeuronReferences[j] = currentForwardLayerNeuron;
			currentForwardLayerNeuron->backNeuronRefConnections[i] = newNeuronConnection;
			currentBackLayerNeuron->frontNeuronRefConnections[j] = newNeuronConnection;

			currentForwardLayerNeuron = currentForwardLayerNeuron->nextNeuron;
		}

	#ifdef ANN
		//create a subnet if necessary
		srand( (unsigned)time(NULL) );	//seeds randomness
		if(currentNumberOfRecursiveSubnets < maxNumberOfRecursiveSubnets)
		{
			double randomNumberBetween0And1 = ((double(abs((short)rand())))/(ABS_OF_SHORT));

			//assume rand() gives double between 0 and 1
			if(randomNumberBetween0And1 > meanSubnetDivergenceFactor)
			{
				Neuron * currentNeuron = currentBackLayerNeuron;
				Neuron * firstNeuronInBackLayerOfSubnet = new Neuron();
				Neuron * currentNeuronInBackLayerOfSubnet = firstNeuronInBackLayerOfSubnet;
				currentNeuron->numNeuronsInBackLayerOfSubnet = currentNeuron->numBackNeuronConnections;
				currentNeuron->numNeuronsInFrontLayerOfSubnet = numberOfNeuronsInNewFrontLayer;
				long q;
				for(q = 0; q < currentNeuron->numBackNeuronConnections; q++)
				{
						//now replace input layer of subnet with currentNeuron->backNeuronReferences [this is not necessary as 1-1 mapping will be assumed here during the neural network training / update algorithm]
							//copyNeuronBPropertiesIntoNeuronA(currentNeuronInBackLayerOfSubnet, currentNeuron->backNeuronReferences[q]);
						//[1-1] map back neurons of currentNeuron with input neurons of subnet		[CHECK THIS - nb this 1-1 mapping must be maintained in the neuronNetworkUpdateAlgorithm, not just in the neuralNetworkFormation]

					Neuron * newNeuron = new Neuron();
					currentNeuronInBackLayerOfSubnet->nextNeuron = newNeuron;
					currentNeuronInBackLayerOfSubnet = currentNeuronInBackLayerOfSubnet->nextNeuron;
				}

				currentNeuron->isSubnet = true;
				currentNeuron->firstNeuronInBackLayerOfSubnet = firstNeuronInBackLayerOfSubnet;
				Neuron * firstNeuronInFrontLayerOfSubnet;
				long meanNumberOfHiddenLayersInSubnet = calcMeanNumberOfHiddenLayersInSubnet(currentNeuron->numNeuronsInBackLayerOfSubnet, currentNeuron->numNeuronsInFrontLayerOfSubnet, meanSubnetDivergenceFactor);
				#define NONLINEARDIVERGENCEFACTOR 0

				fillNondistinctHiddenLayer(firstNeuronInBackLayerOfSubnet, firstNeuronInFrontLayerOfSubnet, currentNeuron->numNeuronsInBackLayerOfSubnet, currentNeuron->numNeuronsInFrontLayerOfSubnet, meanNumberOfHiddenLayersInSubnet, meanSubnetDivergenceFactor+NONLINEARDIVERGENCEFACTOR, maxNumberOfRecursiveSubnets, currentNumberOfRecursiveSubnets+1);
					//NB firstOutputNeuron is not filled here as ANN recursion will not provide a information about the firstOutputNeuronInNetwork

				currentNeuron->firstNeuronInFrontLayerOfSubnet = firstNeuronInFrontLayerOfSubnet;

					//now replace output layer of subnet with frontLayerOfSubnet [this is not necessary as 1-1 mapping will be assumed here during the neural network training / update algorithm]
					//[1-1] map forward layer neurons of currentNeuron with output neurons of subnet	[CHECK THIS - nb this 1-1 mapping must be maintained in the neuronNetworkUpdateAlgorithm, not just in the neuralNetworkFormation]
			}
		}
	#endif

		currentBackLayerNeuron = currentBackLayerNeuron->nextNeuron;
	}



	if(currentNumberOfHiddenLayers == numberOfHiddenLayers)
	{//CHECK THIS - nb this needs to imply that numberOfOutputNeurons = number of neurons in current layer
		if(numberOfOutputNeurons != numberOfNeuronsInNewFrontLayer)
		{
			cout << "Error: error with NN heuristic; (currentNumberOfHiddenLayers == numberOfHiddenLayers) && (numberOfNeuronsInNewTopLevelLayer != numberOfOutputNeurons)" << endl;
			exit(0);
		}

		//reached final output layer
		firstOutputNeuronInNetwork = firstNeuronInNewFrontLayer;
	}
	else
	{
		//now recurse
		firstOutputNeuronInNetwork = fillNondistinctHiddenLayer(firstInputNeuronInNetwork, firstNeuronInNewFrontLayer, numberOfInputNeurons, numberOfOutputNeurons, numberOfNeuronsInNewFrontLayer, (currentNumberOfHiddenLayers+1), numberOfHiddenLayers, meanSubnetDivergenceFactor, maxNumberOfRecursiveSubnets, currentNumberOfRecursiveSubnets);
	}

	return firstOutputNeuronInNetwork;
}

static long calcMeanNumberOfHiddenLayersInSubnet(long numberOfInputNeurons, long numberOfOutputNeurons, double meanSubnetDivergenceFactor)
{
	long answer;
	#ifdef USE_BASIC_SUBNET_DIVERGENCE
		answer = long(double(abs(numberOfInputNeurons - numberOfOutputNeurons))/meanSubnetDivergenceFactor);	//greater the divergence, the less number of hidden layers
	#else
		answer = long(double(abs(numberOfInputNeurons - numberOfOutputNeurons))/meanSubnetDivergenceFactor);	//greater the divergence, the less number of hidden layers
		cout << "error: code is not finished without USE_BASIC_SUBNET_DIVERGENCE" << endl;
	#endif
	return answer;
}


/************************************************************ End Form Advanced Neural Network Routines ******************************************************/















































static long calculateNumberOfNeuronsInNewFrontLayer(long numberOfInputNeurons, long numberOfOutputNeurons, long currentNumberOfHiddenNeurons, long currentNumberOfHiddenLayers, long numberOfHiddenLayers, double meanDivergenceFactor)
{

	long numberOfNeuronsInNewFrontLayer;

#ifdef USE_BASIC_SUBNET_DIVERGENCE

			/*Basic Subnet divergence

				/   On On On  \
			   /   Hn Hn Hn Hn \
			  /  Hn Hn Hn Hn Hn  \
			 /  In In In In In In  \

			using simple layer hidden Neuron heuristic; nb  numberOfHiddenLayers = (numberOfInputNeurons + numberOfOutputNeurons)/2

			*/


	if(numberOfInputNeurons > numberOfOutputNeurons)
	{

		numberOfNeuronsInNewFrontLayer = numberOfInputNeurons + ((numberOfOutputNeurons - numberOfInputNeurons) * (currentNumberOfHiddenLayers/numberOfHiddenLayers));
	}
	else
	{
		numberOfNeuronsInNewFrontLayer = numberOfOutputNeurons + ((numberOfOutputNeurons - numberOfInputNeurons) * (currentNumberOfHiddenLayers/numberOfHiddenLayers));
	}

#elif defined USE_INTERMEDIATE_SUBNET_DIVERGENCE

	long divergenceFactor;


				/*Intermediate Subnet divergence

								/\
							  /    \
							/        \
						  /	           \
						/                \
					  /                    \
					/                        \

				Input 1                        Output 1
				Input 2                        Output 2
				Input 3                        Output 3
				Input 4                        Output 4
				Input N                        Output N

				   \                         /
					 \                     /
					   \                 /
						 \             /
						   \         /
							 \     /
							   \ /



				NB divergence = meanDivergenceFactor until 1/2 meanNumberOfHiddenLayers is reached,  afterwards meanDivergenceFactor = -meanDivergenceFactor  (until number of outputNeurons is reached).
				this implies a ~ diamond shapped (linearily increasing/decreasing) resulting neural net*/

	if(currentNumberOfHiddenLayers < (numberOfHiddenLayers/2))
	{
		divergenceFactor = meanDivergenceFactor;

	}
	else
	{
		divergenceFactor = -meanDivergenceFactor;
	}

	numberOfNeuronsInNewFrontLayer = currentNumberOfHiddenNeurons + currentNumberOfHiddenNeurons * divergenceFactor;

#elif defined USE_ADVANCED_SUBNET_DIVERGENCE

	long divergenceFactor;

				/*
									/\
								   /  \
								  /    \
								/	     \
							  /            \
						   /                  \
						/                        \

					Input 1                        Output 1
					Input 2                        Output 2
					Input 3                        Output 3
					Input 4                        Output 4
					Input N                        Output N

					 \	                         /
						\                     /
						   \               /
							  \          /
								\      /
								 \    /
								  \  /
								   \/

				NB divergence = meanDivergenceFactor*f(currentNumberOfHiddenLayers)  until 1/2 meanNumberOfHiddenLayers is reached,  afterwards meanDivergenceFactor = -meanDivergenceFactor*f(currentNumberOfHiddenLayers)  (until number of outputNeurons is reached)
				depending upon function f this could implies a ~ diamond shapped (non-linearily increasing/decreasing) resulting neural net
				another possibility is that f(currentNumberOfHiddenLayers) could be rand()

			   */

	if(currentNumberOfHiddenLayers < (numberOfHiddenLayers/2))
	{
		divergenceFactor = meanDivergenceFactor * advancedDivergenceFactor(currentNumberOfHiddenLayers, numberOfHiddenLayers);

	}
	else
	{
		divergenceFactor = -meanDivergenceFactor * advancedDivergenceFactor(abs(currentNumberOfHiddenLayers - numberOfHiddenLayers), numberOfHiddenLayers);
	}

	numberOfNeuronsInNewFrontLayer = currentNumberOfHiddenNeurons + currentNumberOfHiddenNeurons * divergenceFactor;

#endif

	return numberOfNeuronsInNewFrontLayer;
}



static double advancedDivergenceFactor(long currentNumberOfHiddenLayersFromNearestEndPoint, long numberOfHiddenLayers)
{
	double advancedDivergenceFactor;

#ifdef USE_ADVANCED_RANDOM_SUBNET_DIVERGENCE
	//seed random number generator
	srand(static_cast<unsigned>(time(0)));
	advancedDivergenceFactor = rand();
#else
	advancedDivergenceFactor = currentNumberOfHiddenLayersFromNearestEndPoint/numberOfHiddenLayers;
#endif

	return advancedDivergenceFactor;
}









