/*******************************************************************************
 *
 * File Name: neuralNetworkTraining.cpp
 * Author: Richard Baxter
 * Project: Advanced Neural Network (ANN) - Fractal Based Neural Network
 * Project Version: 1b13d 29-Sept-07
 * Comments:
 *
 *******************************************************************************/

 //IMPORTANT CODING NOTE - on 10-dec-06/1b6b I have started to remove the neuronReference class as circular referencing does not appear to be allowed in C++
//NB when create NeuronList class change all referecnes to "...List->frontNeuronReferences.." to "...List->neuronReferences" as Neurons when declared for the purpose of creating neuron lists only have their frontNeuronReferences used

#include <ctime>
#include <cstdlib>	//for randon number generation
#include <cmath>

#include <string>
#include <iostream>
#include <fstream>
using namespace std;


#include "neuralNetworkTraining.h"
#include "neuralNetworkUpdateAlgorithm.h"
#include "neuralNetworkFormation.h"
#include "THneuralNetworkParser.h"
#include "THData.h"








/************************************************************ Train Neural Network Routines ******************************************************/



//copy of old c source: this needs to be revised
void THTrainNeuralNetwork(Neuron * firstInputNeuron, Neuron * firstOutputNeuron, int numberOfInputNeurons, int numberOfOutputNeurons, int maxFolds)
{


	/*
	network structure

	X input neurons
	Random hidden neurons with random levels
	Y output neurons

	X = Y
	*/



	//INITIALISES NEURAL NETWORK DATA

	/*
	code has been designed specifically for a single hidden layered neural network
	(with the given input data and its description - see new-thyroid.data/new-thyroid.names)
	*/

	/*
	procedure;

	Go through network 215xaverageNumberOfEpicsx215 times..

	... using each of the 215 cases as a test case, and the other 214 as training cases.

	for each of the 215 test cases;

	train on all 214 cases. then test on the test case. calculate the final error rate.

	average all these error rates over all permutations of test cases to get the final error rate.
	*/

	int foldNum; //number of folds for cross validation

	float totalTestingError = 0;
	float totalTrainingError = 0;
	float totalTrainingEpochs = 0;

	float * trainingEpochsNumberArray = new float[maxFolds];
	float * trainingErrorArray = new float[maxFolds];
	float * testingErrorArray = new float[maxFolds];



	for(foldNum=0; foldNum < maxFolds; foldNum++)
	{

		/*
		assumes Thyroid data is randomly distributed - if its not it should be -
		and takes the last 1/foldNum (*) number of data sets
		as test sets, and the first (foldNum-1)/foldNum as training sets.
		*/

		int numEpochs;
		float testingError;
		float trainingError;

		//train using all sets of data except for test set (= testPermutation)

		resetNeuralNetworkWithRandomBiasAndWeights(firstInputNeuron, numberOfInputNeurons);

		trainingError = 1.0F;
		numEpochs = 0;

		float acceptableError = ACCEPTABLE_ERROR_RATE;
		int maxEpochs = MAX_NUMBER_OF_EPOCHS;


		#ifdef DEBUG
		cout << "q = int(float(maxSetsOfData) * float(foldNum)/float(maxFolds))" << endl;
		cout << "q = " << int(float(maxSetsOfData) * float(foldNum)/float(maxFolds)) << endl;
		cout << "q+int(float(maxSetsOfData)/float(maxFolds)) = " << int(float(maxSetsOfData) * float(foldNum)/float(maxFolds))+int(float(maxSetsOfData)/float(maxFolds)) << endl;
		#endif

		float currentAverageTestingError;
		float recordedTestedError = 1.0F;
			//it is initalised to a high error value so that ... so that overfitting can be identified

		bool overFittingData = false;

		//training
		while((trainingError > acceptableError) && (numEpochs < maxEpochs) && !overFittingData)
		{
			//(*)
			//from start of dataSet -> beginning of test segment
			for(int i = 0; i < int(float(maxSetsOfData) * float(foldNum)/float(maxFolds)); i++)
			{

				//sets inputData into neuralNetwork
				//(NB normalisedInputData[i][0] is target class

				//basic method (from assignment) to mirror in C++
				/*
				outputs0[0] = normalisedInputData[i][1];
				outputs0[1] = normalisedInputData[i][2];
				outputs0[2] = normalisedInputData[i][3];
				outputs0[3] = normalisedInputData[i][4];
				outputs0[4] = normalisedInputData[i][5];

				classTarget[0] = 0.0F;
				classTarget[1] = 0.0F;
				classTarget[2] = 0.0F;
				classTarget[int(normalisedInputData[i][0])] = 1.0F;
				*/
				resetInputsAndClassTargets(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, i);

				trainingError = neuralNetworkBackPropogationPass(firstInputNeuron, firstOutputNeuron);

			}
			//from end of test segment -> end of dataSet
			for(int i = int(float(maxSetsOfData) * float(foldNum)/float(maxFolds))+int(float(maxSetsOfData)/float(maxFolds));i<maxSetsOfData;i++)
			{
				//basic method (from assignment) to mirror in C++
				/*
				outputs0[0] = normalisedInputData[i][1];
				outputs0[1] = normalisedInputData[i][2];
				outputs0[2] = normalisedInputData[i][3];
				outputs0[3] = normalisedInputData[i][4];
				outputs0[4] = normalisedInputData[i][5];

				classTarget[0] = 0.0F;
				classTarget[1] = 0.0F;
				classTarget[2] = 0.0F;
				classTarget[int(normalisedInputData[i][0])] = 1.0F;
				*/
				resetInputsAndClassTargets(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, i);


				//trainingError = backPropogate(numberOfHiddenNeurons, outputs0, outputs1, outputs2, weights1, weights2, biases1, biases2, errors1, errors2, classTarget);
				trainingError = neuralNetworkBackPropogationPass(firstInputNeuron, firstOutputNeuron);
			}

			numEpochs++;

			//testing

			//START OF TESTING FOR TRAINING EPOCH ep

			//creates and stores copy of the trained neural network

			//basic method (from assignment) to mirror in C++
			/*
			for(int j = 0; j < numberOfInputNeurons*numberOfHiddenNeurons; j++)
			{
				weights1Record[j] = weights1[j];
			}
			for(int j = 0; j < numberOfHiddenNeurons*numberOfOutputNeurons; j++)
			{
				weights2Record[j] = weights2[j];
			}
			for(int j = 0; j < numberOfHiddenNeurons; j++)
			{
				biases1Record[j] = biases1[j];
			}
			for(int j = 0; j < numberOfOutputNeurons; j++)
			{
				biases2Record[j] = biases2[j];
			}
			*/
			storeNeuralNetworkBiasAndWeights(firstInputNeuron, numberOfInputNeurons);


			//now determine the number of epochs ep that result in best performance on the respective test partition

			float  sumingTestingError = 0.0F;

			//from start of test segment -> end of test segment
			for(int i = (int(float(maxSetsOfData) * float(foldNum)/float(maxFolds))); i < (int(float(maxSetsOfData) * float(foldNum)/float(maxFolds))+int(float(maxSetsOfData)/float(maxFolds))); i++)
			{
				//basic method (from assignment) to mirror in C++
				/*
				for(j = 0; j < numberOfInputNeurons*numberOfHiddenNeurons; j++)
				{
					 weights1[j] = weights1Record[j];
				}
				for(j = 0; j < numberOfHiddenNeurons*numberOfOutputNeurons; j++)
				{
					weights2[j] = weights2Record[j];
				}
				for(j = 0; j < numberOfHiddenNeurons; j++)
				{
					biases1[j] = biases1Record[j];
				}
				for(j = 0; j < numberOfOutputNeurons; j++)
				{
					biases2[j] = biases2Record[j];
				}
				*/
				restoreNeuralNetworkWithStoredBiasAndWeights(firstInputNeuron, numberOfInputNeurons);

				//basic method (from assignment) to mirror in C++
				/*
				outputs0[0] = normalisedInputData[i][1];
				outputs0[1] = normalisedInputData[i][2];
				outputs0[2] = normalisedInputData[i][3];
				outputs0[3] = normalisedInputData[i][4];
				outputs0[4] = normalisedInputData[i][5];

				classTarget[0] = 0.0F;
				classTarget[1] = 0.0F;
				classTarget[2] = 0.0F;
				classTarget[int(normalisedInputData[i][0])] = 1.0F;
				*/
				resetInputsAndClassTargets(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, i);

				//testingError = backPropogate(numberOfHiddenNeurons, outputs0, outputs1, outputs2, weights1, weights2, biases1, biases2, errors1, errors2, classTarget);
				testingError = neuralNetworkBackPropogationPass(firstInputNeuron, firstOutputNeuron);

				sumingTestingError = sumingTestingError + testingError;
			}

			//resets the trained neural network to the stored copy

			//basic method (from assignment) to mirror in C++
			/*
			for(j = 0; j < numberOfInputNeurons*numberOfHiddenNeurons; j++)
			{
				 weights1[j] = weights1Record[j];
			}
			for(j = 0; j < numberOfHiddenNeurons*numberOfOutputNeurons; j++)
			{
				weights2[j] = weights2Record[j];
			}
			for(j = 0; j < numberOfHiddenNeurons; j++)
			{
				biases1[j] = biases1Record[j];
			}
			for(j = 0; j < numberOfOutputNeurons; j++)
			{
				biases2[j] = biases2Record[j];
			}
			*/
			restoreNeuralNetworkWithStoredBiasAndWeights(firstInputNeuron, numberOfInputNeurons);

			currentAverageTestingError = sumingTestingError/(float(maxSetsOfData)/float(maxFolds));

			//if(testingError > (recordedTestedError+float(ERROR_VARIATION_ALLOWED)))
			if(currentAverageTestingError > (recordedTestedError+float(ERROR_VARIATION_ALLOWED)))
			{
				//Started to overfit data...
				cout << "Overfitting was detected and training was stopped" << endl;
				overFittingData	= true;
			}
			//else if(testingError > recordedTestedError)
			else if(currentAverageTestingError > recordedTestedError)
			{
				//cout << "Potential imminent overfitting..." << endl;
				//do not create a recordedTestedError. Wait to see if there is further testing error increase
			}
			else
			{
				//probably not overfitting the data...

				//recordedTestedError = testingError;
				recordedTestedError = currentAverageTestingError;
			}

			//END OF TESTING FOR TRAINING EPOCH ep


		}

		trainingEpochsNumberArray[foldNum] = numEpochs;
		trainingErrorArray[foldNum] = trainingError;
		//testingErrorArray[foldNum] = testingError;
		testingErrorArray[foldNum] = currentAverageTestingError;


		cout << "Results for fold number = " << foldNum << endl;
		cout << "Error on training Set = " << trainingError << endl;
		cout << "Accuracy of training Set = " << 100.0F*(1.0F-trainingError) << "%" << endl;	//added %s 1-6-04
		cout << "Error on test Set = " <<  currentAverageTestingError << endl;
		cout << "Accuracy of test Set = " << 100.0F*(1.0F- currentAverageTestingError) << "%" << endl;	//added %s 1-6-04
		cout << "Number of epochs = " << numEpochs << "\n" << endl;

		totalTrainingEpochs = totalTrainingEpochs + numEpochs;
		totalTrainingError = totalTrainingError + trainingError;
		//totalTestingError = totalTestingError + testingError;
		totalTestingError = totalTestingError + currentAverageTestingError;

	}


	//calculates and prints out average error;

	//calculate averages;
	float averageTrainingError = totalTrainingError/float(maxFolds);
	float averageTestingError = totalTestingError/float(maxFolds);
	float averageTotalTrainingEpochs = totalTrainingEpochs/float(maxFolds);

	//calculate standard deviations;
	float sdTrainingError = calculateStandardDeviation(trainingErrorArray, maxFolds, averageTrainingError);
	float sdTestingError = calculateStandardDeviation(testingErrorArray, maxFolds, averageTestingError);
	float sdTotalTrainingEpochs = calculateStandardDeviation(trainingEpochsNumberArray, maxFolds, averageTotalTrainingEpochs);

	//time feedback (used to calculate the optimum number of hidden neurons
	/*
	struct tm *myTime;
	time_t myClock = time(NULL);
	char buf[BUFSIZ];
	myTime = localtime(&myClock);
	*/
	//cout << "time = " << asctime(myTime) << endl;


	cout << "\n*****************************************************" << endl;
	cout << "Average Training Error = " << averageTrainingError << "\n" <<endl;
	cout << "Average Training Accuracy = " << 100.0F*(1.0F-averageTrainingError) << "%" << "\n" <<endl;	//added %s 1-6-04
	cout << "Average Testing Error = " << averageTestingError << "\n" <<endl;
	cout << "Average Testing Accuracy = " << 100.0F*(1.0F-averageTestingError) << "%" << "\n" <<endl;
	cout << "Average number of training Epochs = " << averageTotalTrainingEpochs << "\n" <<endl;	//added %s 1-6-04
	cout << "Standard Deviation Training Error = " << sdTrainingError << "\n" <<endl;
	cout << "Standard Deviation Testing Error = " << sdTestingError << "\n" <<endl;
	cout << "Standard Deviation of number of training Epochs = " << sdTotalTrainingEpochs << "\n" <<endl;


	//now train the network on all examples for the mean number of epochs across all folds
	resetNeuralNetworkWithRandomBiasAndWeights(firstInputNeuron, numberOfInputNeurons);

	float finalError;
	for(int e=0;e < averageTotalTrainingEpochs;e++)
	{

		//from start of dataSet -> end of dataSet
		for(int i = 0; i < maxSetsOfData; i++)
		{

			//basic method (from assignment) to mirror in C++
			/*
			outputs0[0] = normalisedInputData[i][1];
			outputs0[1] = normalisedInputData[i][2];
			outputs0[2] = normalisedInputData[i][3];
			outputs0[3] = normalisedInputData[i][4];
			outputs0[4] = normalisedInputData[i][5];

			classTarget[0] = 0.0F;
			classTarget[1] = 0.0F;
			classTarget[2] = 0.0F;
			classTarget[int(normalisedInputData[i][0])] = 1.0F;
			*/
			resetInputsAndClassTargets(firstInputNeuron, firstOutputNeuron, numberOfInputNeurons, numberOfOutputNeurons, i);

			//finalError = backPropogate(numberOfHiddenNeurons, outputs0, outputs1, outputs2, weights1, weights2, biases1, biases2, errors1, errors2, classTarget);
			finalError = neuralNetworkBackPropogationPass(firstInputNeuron, firstOutputNeuron);
		}
	}
	cout << "Final accuracy for e epochs (based on results from 10 fold cross validation) = " << (1.0F-finalError)*100.0F << "%" << "\n" <<endl;
	cout << "***************************************************** \n " << endl;

}



/************************************************************ End Neural Network Routines ******************************************************/



























/************************************************************ Store/Record Neural Network weights and biases Routines ******************************************************/


void storeNeuralNetworkBiasAndWeights(Neuron * firstInputNeuron, long numberOfInputNeurons)
{
	Neuron * currentNeuronReference = firstInputNeuron;

	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		#ifdef DEBUG
		cout << "\nA low level Input neuron has been selected for Reseting the Neural Network with previously stored Biases And Weights" << endl;
		cout << "neuron ID = " << currentNeuronReference->id << endl;
		cout << "neuron Bias = " << currentNeuronReference->bias << endl;
		#endif

		storeNeuronParentsBiasAndWeight(currentNeuronReference);
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}

	//basic restore neural network with stored biases and weights algorithm (from assignment) to mirror in C++
	/*
	for(int j = 0; j < numberOfInputNeurons*numberOfHiddenNeurons; j++)
	{
		weights1Record[j] = weights1[j];
	}
	for(int j = 0; j < numberOfHiddenNeurons*numberOfOutputNeurons; j++)
	{
		weights2Record[j] = weights2[j];
	}
	for(int j = 0; j < numberOfHiddenNeurons; j++)
	{
		biases1Record[j] = biases1[j];
	}
	for(int j = 0; j < numberOfOutputNeurons; j++)
	{
		biases2Record[j] = biases2[j];
	}
	*/


}

		//UP TO HERE - realise there is a problem, need to separate neuron connnections from neurons themselves...

void storeNeuronParentsBiasAndWeight(Neuron * neuronBeingAccessed)
{
	for(long i = 0; i < neuronBeingAccessed->numFrontNeuronConnections; i++)
	{
		#ifdef DEBUG
		cout << "neuron Weight = " << neuronBeingAccessed->frontNeuronRefConnections[i]->weight << endl;
		#endif

		neuronBeingAccessed->frontNeuronRefConnectionsRecord[i]->weight = neuronBeingAccessed->frontNeuronRefConnections[i]->weight;
		neuronBeingAccessed->frontNeuronReferences[i]->storedBias = neuronBeingAccessed->frontNeuronReferences[i]->bias;

		//recursion [NB there will be unnecessay redundancy here, as the same neuron will be accessed more than once]*/
		if(neuronBeingAccessed->frontNeuronReferences[i]->numFrontNeuronConnections != 0)
		{
			storeNeuronParentsBiasAndWeight(neuronBeingAccessed->frontNeuronReferences[i]);
		}

	#ifdef ANN
		if(neuronBeingAccessed->isSubnet)
		{
			storeNeuralNetworkBiasAndWeights(neuronBeingAccessed->firstNeuronInBackLayerOfSubnet, neuronBeingAccessed->numNeuronsInBackLayerOfSubnet);
		}
	#endif
	}
}
/************************************************************ End Store/Record Neural Network weights and biases Routines **************************************************/




/************************************************************ Restore Neural Network with previously recorded weights and biases Routines ******************************************************/


void restoreNeuralNetworkWithStoredBiasAndWeights(Neuron * firstInputNeuron, long numberOfInputNeurons)
{
	Neuron * currentNeuronReference = firstInputNeuron;

	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		#ifdef DEBUG
		cout << "\nA low level Input neuron has been selected for Reseting the Neural Network with previously stored Biases And Weights" << endl;
		cout << "neuron ID = " << firstInputNeuron->id << endl;
		cout << "neuron Bias = " << firstInputNeuron->bias << endl;
		#endif

		restoreNeuronParentsWithStoredBiasAndWeight(currentNeuronReference);
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}

	//basic restore neural network with stored biases and weights algorithm (from assignment) to mirror in C++
	/*
	//resets the trained neural network to the stored copy
	for(j = 0; j < numberOfInputNeurons*numberOfHiddenNeurons; j++)
	{
		 weights1[j] = weights1Record[j];
	}
	for(j = 0; j < numberOfHiddenNeurons*numberOfOutputNeurons; j++)
	{
		weights2[j] = weights2Record[j];
	}
	for(j = 0; j < numberOfHiddenNeurons; j++)
	{
		biases1[j] = biases1Record[j];
	}
	for(j = 0; j < numberOfOutputNeurons; j++)
	{
		biases2[j] = biases2Record[j];
	}
	*/

}

void restoreNeuronParentsWithStoredBiasAndWeight(Neuron * neuronBeingAccessed)
{
	for(long i = 0; i < neuronBeingAccessed->numFrontNeuronConnections; i++)
	{
		#ifdef DEBUG
		cout << "neuron Weight = " << neuronBeingAccessed->frontNeuronRefConnections[i]->weight << endl;
		#endif

		neuronBeingAccessed->frontNeuronRefConnections[i]->weight = neuronBeingAccessed->frontNeuronRefConnectionsRecord[i]->weight;
		neuronBeingAccessed->frontNeuronReferences[i]->bias = neuronBeingAccessed->frontNeuronReferences[i]->storedBias;

		//recursion [NB there will be unnecessay redundancy here, as the same neuron will be accessed more than once]*/
		if(neuronBeingAccessed->frontNeuronReferences[i]->numFrontNeuronConnections != 0)
		{
			restoreNeuronParentsWithStoredBiasAndWeight(neuronBeingAccessed->frontNeuronReferences[i]);
		}

	#ifdef ANN
		if(neuronBeingAccessed->isSubnet)
		{
			restoreNeuralNetworkWithStoredBiasAndWeights(neuronBeingAccessed->firstNeuronInBackLayerOfSubnet, neuronBeingAccessed->numNeuronsInBackLayerOfSubnet);
		}
	#endif
	}
}
/************************************************************ End Restore Neural Network with previously recorded weights and biases Routines **************************************************/




/************************************************************ Reset Neural Network with random biases and weights Routines ******************************************************/
void resetNeuralNetworkWithRandomBiasAndWeights(Neuron * firstInputNeuron, long numberOfInputNeurons)
{
	srand(static_cast<unsigned>(time(0)));		//seed random number generator

	Neuron * currentNeuronReference = firstInputNeuron;
	for(long i = 0; i < numberOfInputNeurons; i++)
	{

		#ifdef DEBUG
		cout << "\nA low level Input neuron has been selected for Reseting the Neural Network with Random Biases And Weights" << endl;
		cout << "neuron ID = " << firstInputNeuron->frontNeuronReferences[i]->id << endl;
		cout << "neuron Bias = " << firstInputNeuron->frontNeuronReferences[i]->bias << endl;
		#endif

		resetNeuronParentsWithRandomBiasAndWeight(currentNeuronReference);
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}


	//basic reset neurons with random biases and weights algorithm (from assignment) to mirror in C++
	/*
	int i, j;
	for(i = 0; i < numberOfInputNeurons; i++)
	{
		for(j = 0; j < numberOfHiddenNeurons; j++)
		{
				//creates a random float between -1 and 1
			weights1[i*numberOfHiddenNeurons+j] = (float(rand() * 2.0)/(float(RAND_MAX)+1.0))-1.0;
			//weights between input and hidden layer
		}
	}
	for(i = 0; i < numberOfHiddenNeurons; i++)
	{
		for(j = 0; j < numberOfOutputNeurons; j++)
		{
			weights2[i*numberOfOutputNeurons+j] = (float(rand() * 2.0)/(float(RAND_MAX)+1.0))-1.0;
			//weights between hidden layer and outputlayer
		}
	}
	for(i = 0; i < numberOfHiddenNeurons; i++)
	{
		biases1[i] = (float(rand() * 2.0)/(float(RAND_MAX)+1.0))-1.0;
		//biases of hidden neurons
	}
	for(i = 0; i < numberOfOutputNeurons; i++)
	{
		biases2[i] = (float(rand() * 2.0)/(float(RAND_MAX)+1.0))-1.0;
		//biases of output neurons
	}
	*/

	/*
	ARRAYS NOT FILLED/RESET HERE

	//this are over-riden every backpropagation and so they do not have to be initalised;


	outputs1[numberOfHiddenNeurons];
			//outputs values of hiddenLayer
	outputs2[numberOfOutputNeurons];
			//outputs values of outputLayer
	errors1[numberOfHiddenNeurons]
			//error values of hiddenLayer
	errors2[numberOfOutputNeurons]
			//error values of outputLayer

	//these are filled in elsewhere;

	outputs0[numberOfHiddenNeurons];
			//outputs values of inputLayer
	classTarget[numberOfOutputNeurons];

	*/

}


void resetNeuronParentsWithRandomBiasAndWeight(Neuron * neuronBeingAccessed)
{
	for(long i = 0; i < neuronBeingAccessed->numFrontNeuronConnections; i++)
	{

		#ifdef DEBUG
		cout << "neuron Weight = " << neuronBeingAccessed->frontNeuronRefConnections[i]->weight << endl;
		#endif

		neuronBeingAccessed->frontNeuronRefConnections[i]->weight = (double(rand() * 2.0F)/(double(RAND_MAX)+1.0F))-1.0F;
		neuronBeingAccessed->frontNeuronReferences[i]->bias = (double(rand() * 2.0F)/(double(RAND_MAX)+1.0F))-1.0F;

		//recursion
		if(neuronBeingAccessed->frontNeuronReferences[i]->numFrontNeuronConnections != 0)
		{
			resetNeuronParentsWithRandomBiasAndWeight(neuronBeingAccessed->frontNeuronReferences[i]);
		}
	}
}
/************************************************************ End Reset Neural Network with random biases and weights Routines **************************************************/





/************************************************************ Reset Input/ClassTarget Neurons Neural Network Routines ******************************************************/
void resetInputsAndClassTargets(Neuron * firstInputNeuron, Neuron * firstOutputNeuron, long numberOfInputNeurons, long numberOfOutputNeurons, int testSegment)
{

	//sets inputData into neuralNetwork
	//(NB normalisedInputData[i][0] is target class

	//basic method (from assignment) to mirror in C++
	/*
	outputs0[0] = normalisedInputData[i][1];
	outputs0[1] = normalisedInputData[i][2];
	outputs0[2] = normalisedInputData[i][3];
	outputs0[3] = normalisedInputData[i][4];
	outputs0[4] = normalisedInputData[i][5];
	*/


	Neuron * currentNeuronReference = firstInputNeuron;
	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		currentNeuronReference->output = normalisedInputData[testSegment][i];
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}

	//basic method (from assignment) to mirror in C++
	/*
	classTarget[0] = 0.0F;
	classTarget[1] = 0.0F;
	classTarget[2] = 0.0F;
	classTarget[int(normalisedInputData[i][0])] = 1.0F;
	*/

	currentNeuronReference = firstOutputNeuron;
	for(long i = 0; i < numberOfOutputNeurons; i++)
	{
		currentNeuronReference->classTarget = 0.0F;

		if(i == normalisedInputData[testSegment][0])
		{
			currentNeuronReference->classTarget = 1.0F;
		}

		currentNeuronReference = currentNeuronReference->nextNeuron;
	}

#ifdef ANN
	/*CHECK THIS: anything required for subnet inputs/outputs???*/
#endif

}

/************************************************************ End Reset Input/ClassTarget Neurons Neural Network Routines ******************************************************/



float calculateStandardDeviation(float* array, int length, float average)
{
	float sum = 0;

	for(int i=0;i<length;i++)
	{
		sum = sum + pow((array[i]-average),2);
	}

	float sd = sqrt(sum/float(length));

	return sd;
}
