/*******************************************************************************
 *
 * File Name: neuronClass.h
 * Author: Richard Baxter
 * Project: Advanced Neural Network (ANN) - Fractal Based Neural Network
 * Project Version: 1b13d 29-Sept-07
 * Comments:
 *
 *******************************************************************************/

 //IMPORTANT CODING NOTE - on 10-dec-06/1b6b I have started to remove the neuronReference class as circular referencing does not appear to be allowed in C++
//NB when create NeuronList class change all referecnes to "...List->frontNeuronReferences.." to "...List->neuronReferences"

/************************************************************ Neural Network Class Definitions ******************************************************/


#ifndef NEURON_CLASS_HEADER
#define NEURON_CLASS_HEADER

#define MAX_NUMBER_OF_NEURON_CONNECTIONS 100


//#define TEST_LEVEL_0_TH			//emulate original C Code - nn 1aFinal with optimum number of hidden neurons
#define TEST_LEVEL_1_TH				//emulate original C Code - nn 1aFinal with range of hidden neurons
//#define TEST_LEVEL_2_TH
//#define TEST_LEVEL_3_TH			//with range of nn layers
//#define TEST_LEVEL_4_TH
//#define TEST_LEVEL_5_TH			//with range of nn subnets
#ifdef TEST_LEVEL_4_TH
	#define ANN
#elif defined TEST_LEVEL_5_TH
	#define ANN
#endif



class NeuronConnection
{
public:

	NeuronConnection(void);
	~NeuronConnection(void);

	double weight;
};

	//needs to be updated with ANN spec;
class Neuron
{
public:

	Neuron(void);
	~Neuron(void);

	//NB max number of neurons in a neural network is currently 2^64
	//NB 64 bit precision is used;
	long id;

	double bias;
	double storedBias;					// (storedBias used for reset)
	double output;
	double classTarget;					//only relevant to output layer neurons
	double error;	//added 28 Sept 07

	//neuron relationships

		//whole layer relationships
	bool hasFrontLayer;	//new - required
	bool hasBackLayer;	//new - required
	Neuron * firstNeuronInFrontLayer;	//new - required
	Neuron * firstNeuronInBackLayer;	//new - required
	Neuron * nextNeuron;		/*next neuron on same layer*/
	Neuron * previousNeuron;	/*previous neuron on same layer*/

		//partial layer relationships [NB all neurons on layer a do not have to be connected to all neurons on layer b]
	long numFrontNeuronConnections;
	long numBackNeuronConnections;
	Neuron * frontNeuronReferences[MAX_NUMBER_OF_NEURON_CONNECTIONS];		//an array of pointers to neurons
	Neuron * backNeuronReferences[MAX_NUMBER_OF_NEURON_CONNECTIONS];		//an array of pointers to neurons		[used for nn navigation only]
	NeuronConnection * frontNeuronRefConnections[MAX_NUMBER_OF_NEURON_CONNECTIONS];			//an array of pointers to neuron connections
	NeuronConnection * frontNeuronRefConnectionsRecord[MAX_NUMBER_OF_NEURON_CONNECTIONS];	//an array of pointers to neuron connections		// (frontNeuronRefConnectionWeightsRecord used for reset)
	NeuronConnection * backNeuronRefConnections[MAX_NUMBER_OF_NEURON_CONNECTIONS];			//an array of pointers to neuron connections
	NeuronConnection * backNeuronRefConnectionsRecord[MAX_NUMBER_OF_NEURON_CONNECTIONS];	//an array of pointers to neuron connections		// (backNeuronRefConnectionWeightsRecord used for reset)

		//NB there is a 1-1 mapping between neuron references and weights [would use a new class called NeuronReference which contained Neuron *, weight, and nextNeuron *, but circular references in C++ are not allowed
			//NB there also is a 1-1 mapping between child neuron front weights and parent neuron back weights

	//old terms;
	//long currentFrontNeuronReference;	//currently only useful/used during nn creation/formation
	//long currentBackNeuronReference;	//currently only useful/used during nn creation/formation
	//frontBackNeuronReference
	//firstBackNeuronReference

#ifdef ANN

	bool isSubnet;	 	//alternative to being a subnet is being a distinct neuron

	//variables only used by subnets (abstract neurons)
	Neuron * firstNeuronInBackLayerOfSubnet;	//this variable only used if this object is a subnet, not a neuron - the properties of the neurons in this list are direct copies of those in firstNeuronInFrontLayer
	Neuron * firstNeuronInFrontLayerOfSubnet;

	long numNeuronsInBackLayerOfSubnet;
	long numNeuronsInFrontLayerOfSubnet;


#endif

};

void copyNeuronKeyProperties(Neuron * neuronToUpdate, Neuron * neuronToCopy);

#endif

/************************************************************ End Neural Network Class Definitions **************************************************/

